/*
	**************************************************************************
	(C) 2012. IntBuSoft Ltd. Aztec Marker Encode Decode. http://intbusoft.com
	adcoding.cpp
	**************************************************************************
	Programer: Kruchinin Aleksandr.
	**************************************************************************
	May be used under the terms of the GNU Lesser General Public License (LGPL)   
*/

#include "rs_export.h"
#include <memory.h>
#include "adcoding.h"

#ifdef TYPE_AM
// Template of marker
bool Matrix[121]={
	1,1,0,0,0,0,0,0,0,0,1,
	1,1,1,1,1,1,1,1,1,1,1,
	0,1,0,0,0,0,0,0,0,1,0,
	0,1,0,1,1,1,1,1,0,1,0,
	0,1,0,1,0,0,0,1,0,1,0,
	0,1,0,1,0,0,0,1,0,1,0,
	0,1,0,1,0,0,0,1,0,1,0,
	0,1,0,1,1,1,1,1,0,1,0,
	0,1,0,0,0,0,0,0,0,1,0,
	0,1,1,1,1,1,1,1,1,1,1,
	0,0,0,0,0,0,0,0,0,0,0
};
#else
bool Matrix[100]={
	1,1,0,0,0,0,0,0,0,1,
	1,1,1,1,1,1,1,1,1,1,
	0,1,0,0,0,0,0,0,1,0,
	0,1,0,1,1,1,1,0,1,0,
	0,1,0,1,0,0,1,0,1,0,
	0,1,0,1,0,0,1,0,1,0,
	0,1,0,1,1,1,1,0,1,0,
	0,1,0,0,0,0,0,0,1,0,
	0,1,1,1,1,1,1,1,1,1,
	0,0,0,0,0,0,0,0,0,0
};
#endif

#define RS_SYMSZ 4
#define RS_N (1 << RS_SYMSZ) - 1 

/*
	Aztec Marker data:
	8 bit - data (2 - dw);
	16 bit - correct Reed-Solomon (4 - rsw).
*/

void EncodeAM_(unsigned char *symb,unsigned char* symbrs)
{
#ifdef TYPE_AM
	void *rs = init_rs_char(RS_SYMSZ, 19, 1, 1, 5, RS_N - 7);
#else
	void *rs = init_rs_char(RS_SYMSZ, 19, 1, 1, 4, RS_N - 6);
#endif
	encode_rs_char(rs, symb, symbrs);
	free_rs_char((void*)rs);
}

int DecodeAM_(unsigned char* symbrs)
{
#ifdef TYPE_AM
	void *rs = init_rs_char(RS_SYMSZ, 19, 1, 1, 5, RS_N - 7);	
#else
	void *rs = init_rs_char(RS_SYMSZ, 19, 1, 1, 4, RS_N - 6);	
#endif
	int err= decode_rs_char(rs,symbrs, 0, 0);
	free_rs_char((void*)rs);
	return err;
}



void EncodeAM(unsigned char symb,bool* out)
{
	memcpy(out,Matrix,sizeof(Matrix));

	unsigned char b[7];
	b[0]=(symb&0xF0) >> 4;
	b[1]=(symb&0x0F);

	EncodeAM_(&b[0],&b[2]);

	// Modify template
#ifdef TYPE_AM
	for(int i=0;i<28;i++)
	{
		if (i<7) out[2+i]=( b[int(i/4)] & ( 1 << (3-(i%4))) ) > 0;
		else{
			if (i<14) out[10+(2+i-7)*11]=( b[int(i/4)] & ( 1 << (3-(i%4))) ) > 0;
			else{
				if (i<21) out[8-(i-14)+110]=( b[int(i/4)] & ( 1 << (3-(i%4))) ) > 0;
				else out[(8-(i-21))*11]=( b[int(i/4)] & ( 1 << (3-(i%4))) ) > 0;
			}
		}
	}
#else
	for(int i=0;i<24;i++)
	{
		if (i<6) out[2+i]=( b[int(i/4)] & ( 1 << (3-(i%4))) ) > 0;
		else{
			if (i<12) out[9+(2+i-6)*10]=( b[int(i/4)] & ( 1 << (3-(i%4))) ) > 0;
			else{
				if (i<18) out[7-(i-12)+90]=( b[int(i/4)] & ( 1 << (3-(i%4))) ) > 0;
				else out[(7-(i-18))*10]=( b[int(i/4)] & ( 1 << (3-(i%4))) ) > 0;
			}
		}
	}
#endif
}

int DecodeAM(bool* in,unsigned char*out)
{	
	unsigned char b[7];
	memset(b,0,7);

	// Read Data
#ifdef TYPE_AM
	for(int i=0;i<28;i++)
	{
		if (i<7)  b[int(i/4)]+= in[2+i]*( 1 << (3-(i%4)));
		else{
			if (i<14)  b[int(i/4)]+= in[10+(2+i-7)*11]*( 1 << (3-(i%4)));
			else{
				if (i<21) b[int(i/4)]+= in[8-(i-14)+110]*( 1 << (3-(i%4)));
				else b[int(i/4)]+= in[(8-(i-21))*11]*( 1 << (3-(i%4)));
			}
		}
	}
#else
	for(int i=0;i<24;i++)
	{
		if (i<6)  b[int(i/4)]+= in[2+i]*( 1 << (3-(i%4)));
		else{
			if (i<12)  b[int(i/4)]+= in[9+(2+i-6)*10]*( 1 << (3-(i%4)));
			else{
				if (i<18 ) b[int(i/4)]+= in[7-(i-12)+90]*( 1 << (3-(i%4)));
				else b[int(i/4)]+= in[(7-(i-18))*10]*( 1 << (3-(i%4)));
			}
		}
	}
#endif

	int err=DecodeAM_(b);

	(*out)=(b[0]<<4)+b[1];
	return err;
}